@@ -898,11 +898,11 @@ urls_have_same_origin (const char *a_url,
   g_autoptr (GUri) a_uri = NULL;
   g_autoptr (GUri) b_uri = NULL;
 
-  a_uri = g_uri_parse (a_url, G_URI_FLAGS_NONE, NULL);
+  a_uri = g_uri_parse (a_url, G_URI_FLAGS_PARSE_RELAXED, NULL);
   if (!a_uri || !g_uri_get_host (a_uri))
     return FALSE;
 
-  b_uri = g_uri_parse (b_url, G_URI_FLAGS_NONE, NULL);
+  b_uri = g_uri_parse (b_url, G_URI_FLAGS_PARSE_RELAXED, NULL);
   if (!b_uri || !g_uri_get_host (b_uri))
     return FALSE;
 
@@ -969,62 +969,69 @@ ephy_web_icon_copy_cb (GFile        *file,
 gboolean
 ephy_web_application_save (EphyWebApplication *app)
 {
+  g_autoptr (GKeyFile) keyfile = NULL;
   g_autofree char *contents = NULL;
+  g_autofree char *name = NULL;
+  g_autofree char *icon = NULL;
+  g_autofree char *exec = NULL;
+  g_auto (GStrv) strings = NULL;
+  guint exec_length;
+  gboolean changed = FALSE;
   gboolean saved = FALSE;
   g_autoptr (GError) error = NULL;
 
   g_assert (!ephy_is_running_inside_sandbox ());
 
-  if (g_file_get_contents (app->desktop_path, &contents, NULL, &error)) {
-    g_autoptr (GKeyFile) key = NULL;
-    g_autofree char *name = NULL;
-    g_autofree char *icon = NULL;
-    g_autofree char *exec = NULL;
-    g_auto (GStrv) strings = NULL;
-    guint exec_length;
-    gboolean changed = FALSE;
-
-    key = g_key_file_new ();
-    g_key_file_load_from_data (key, contents, -1, 0, NULL);
-
-    name = g_key_file_get_string (key, "Desktop Entry", "Name", NULL);
-    if (g_strcmp0 (name, app->name) != 0) {
-      changed = TRUE;
-      g_key_file_set_string (key, "Desktop Entry", "Name", app->name);
-    }
-
-    icon = g_key_file_get_string (key, "Desktop Entry", "Icon", NULL);
-    if (g_strcmp0 (icon, app->icon_path) != 0) {
-      g_autoptr (GFile) new_icon = NULL;
-      g_autoptr (GFile) old_icon = NULL;
-      changed = TRUE;
-      new_icon = g_file_new_for_path (app->icon_path);
-      old_icon = g_file_new_for_path (icon);
-      g_file_copy_async (new_icon, old_icon, G_FILE_COPY_OVERWRITE,
-                         G_PRIORITY_DEFAULT, NULL, NULL, NULL,
-                         (GAsyncReadyCallback)ephy_web_icon_copy_cb, NULL);
-    }
-
-    exec = g_key_file_get_string (key, "Desktop Entry", "Exec", NULL);
-    strings = g_strsplit (exec, " ", -1);
-
-    exec_length = g_strv_length (strings);
-    if (g_strcmp0 (strings[exec_length - 1], app->url) != 0) {
-      changed = TRUE;
-      g_free (strings[exec_length - 1]);
-      strings[exec_length - 1] = g_strdup (app->url);
-      g_free (exec);
-      exec = g_strjoinv (" ", strings);
-      g_key_file_set_string (key, "Desktop Entry", "Exec", exec);
-    }
-
-    if (changed) {
-      saved = g_key_file_save_to_file (key, app->desktop_path, &error);
-      if (!saved)
-        g_warning ("Failed to save desktop file of web application: %s\n", error->message);
-    }
-  } else {
+  if (!g_file_get_contents (app->desktop_path, &contents, NULL, &error)) {
     g_warning ("Failed to load desktop file of web application: %s\n", error->message);
+    return FALSE;
+  }
+
+  keyfile = g_key_file_new ();
+  g_key_file_load_from_data (keyfile, contents, -1, 0, NULL);
+
+  name = g_key_file_get_string (keyfile, "Desktop Entry", "Name", NULL);
+  if (g_strcmp0 (name, app->name) != 0) {
+    changed = TRUE;
+    g_key_file_set_string (keyfile, "Desktop Entry", "Name", app->name);
+  }
+
+  icon = g_key_file_get_string (keyfile, "Desktop Entry", "Icon", NULL);
+  if (g_strcmp0 (icon, app->icon_path) != 0) {
+    g_autoptr (GFile) new_icon = NULL;
+    g_autoptr (GFile) old_icon = NULL;
+    changed = TRUE;
+    new_icon = g_file_new_for_path (app->icon_path);
+    old_icon = g_file_new_for_path (icon);
+    g_file_copy_async (new_icon, old_icon, G_FILE_COPY_OVERWRITE,
+                       G_PRIORITY_DEFAULT, NULL, NULL, NULL,
+                       (GAsyncReadyCallback)ephy_web_icon_copy_cb, NULL);
+  }
+
+  exec = g_key_file_get_string (keyfile, "Desktop Entry", "Exec", NULL);
+  strings = g_strsplit (exec, " ", -1);
+
+  exec_length = g_strv_length (strings);
+  if (g_strcmp0 (strings[exec_length - 1], app->url) != 0) {
+    changed = TRUE;
+    g_free (strings[exec_length - 1]);
+    strings[exec_length - 1] = g_strdup (app->url);
+    g_free (exec);
+    exec = g_strjoinv (" ", strings);
+    g_key_file_set_string (keyfile, "Desktop Entry", "Exec", exec);
+  }
+
+  if (changed) {
+    char *resolved_path = realpath (app->desktop_path, NULL);
+    if (!resolved_path) {
+      g_warning ("Failed to save web application %s: failed to resolve path %s: %s", app->name, app->desktop_path, g_strerror (errno));
+      return FALSE;
+    }
+
+    saved = g_key_file_save_to_file (keyfile, resolved_path, &error);
+    if (!saved)
+      g_warning ("Failed to save web application %s desktop file %s: %s\n", app->name, resolved_path, error->message);
+    free (resolved_path);
   }
 
   return saved;
