@@ -26,6 +26,7 @@
 #include "ephy-bookmark-row.h"
 #include "ephy-bookmarks-manager.h"
 #include "ephy-debug.h"
+#include "ephy-link.h"
 #include "ephy-shell.h"
 #include "ephy-window.h"
 
@@ -44,33 +45,46 @@ struct _EphyBookmarksPopover {
   EphyBookmarksManager *manager;
 };
 
-G_DEFINE_TYPE (EphyBookmarksPopover, ephy_bookmarks_popover, GTK_TYPE_POPOVER)
+G_DEFINE_FINAL_TYPE (EphyBookmarksPopover, ephy_bookmarks_popover, GTK_TYPE_POPOVER)
 
 #define EPHY_LIST_BOX_ROW_TYPE_BOOKMARK "bookmark"
 #define EPHY_LIST_BOX_ROW_TYPE_TAG "tag"
 
-static GtkWidget * create_bookmark_row (gpointer item, gpointer user_data);
+static GtkWidget *create_bookmark_row (gpointer item, gpointer user_data);
 static GtkWidget *create_tag_row (const char *tag);
 
 static void
-remove_bookmark_row_from_container (GtkContainer *container,
-                                    const char   *url)
+tag_detail_back (EphyBookmarksPopover *self)
 {
-  GList *children;
+  GtkListBoxRow *row;
 
-  g_assert (GTK_IS_CONTAINER (container));
+  g_assert (EPHY_IS_BOOKMARKS_POPOVER (self));
 
-  children = gtk_container_get_children (container);
-  for (GList *l = children; l && l->data; l = l->next) {
-    const char *type = g_object_get_data (l->data, "type");
+  gtk_stack_set_visible_child_name (GTK_STACK (self->toplevel_stack),
+                                    "default");
+
+  while ((row = gtk_list_box_get_row_at_index (GTK_LIST_BOX (self->tag_detail_list_box), 0)))
+    gtk_list_box_remove (GTK_LIST_BOX (self->tag_detail_list_box), GTK_WIDGET (row));
+}
+
+static void
+remove_bookmark_row (GtkListBox *list_box,
+                     const char *url)
+{
+  GtkListBoxRow *row;
+  int i = 0;
+
+  g_assert (GTK_IS_LIST_BOX (list_box));
+
+  while ((row = gtk_list_box_get_row_at_index (list_box, i++))) {
+    const char *type = g_object_get_data (G_OBJECT (row), "type");
 
     if (g_strcmp0 (type, EPHY_LIST_BOX_ROW_TYPE_BOOKMARK) == 0 &&
-        g_strcmp0 (ephy_bookmark_row_get_bookmark_url (l->data), url) == 0) {
-      gtk_container_remove (container, l->data);
+        g_strcmp0 (ephy_bookmark_row_get_bookmark_url (EPHY_BOOKMARK_ROW (row)), url) == 0) {
+      gtk_list_box_remove (list_box, GTK_WIDGET (row));
       break;
     }
   }
-  g_list_free (children);
 }
 
 static void
@@ -79,21 +93,18 @@ ephy_bookmarks_popover_bookmark_tag_added_cb (EphyBookmarksPopover *self,
                                               const char           *tag,
                                               EphyBookmarksManager *manager)
 {
-  GtkWidget *tag_row;
-  GList *children;
-  GList *l;
   gboolean exists;
   const char *visible_stack_child;
-  const char *title;
-  const char *type;
+  GtkListBoxRow *row;
+  int i = 0;
 
   g_assert (EPHY_IS_BOOKMARK (bookmark));
   g_assert (EPHY_IS_BOOKMARKS_POPOVER (self));
 
   /* If the bookmark no longer has 0 tags, we remove it from the tags list box */
   if (g_sequence_get_length (ephy_bookmark_get_tags (bookmark)) == 1)
-    remove_bookmark_row_from_container (GTK_CONTAINER (self->tags_list_box),
-                                        ephy_bookmark_get_url (bookmark));
+    remove_bookmark_row (GTK_LIST_BOX (self->tags_list_box),
+                         ephy_bookmark_get_url (bookmark));
 
   /* If we are on the tag detail list box, then the user has toggled the state
    * of the tag widget multiple times. The first time the bookmark was removed
@@ -101,17 +112,17 @@ ephy_bookmarks_popover_bookmark_tag_added_cb (EphyBookmarksPopover *self,
   visible_stack_child = gtk_stack_get_visible_child_name (GTK_STACK (self->toplevel_stack));
   if (g_strcmp0 (visible_stack_child, "tag_detail") == 0 &&
       g_strcmp0 (self->tag_detail_tag, tag) == 0) {
-    GtkWidget *row;
+    GtkWidget *bookmark_row;
 
-    row = create_bookmark_row (bookmark, self);
-    gtk_container_add (GTK_CONTAINER (self->tag_detail_list_box), row);
+    bookmark_row = create_bookmark_row (bookmark, self);
+    gtk_list_box_append (GTK_LIST_BOX (self->tag_detail_list_box), bookmark_row);
   }
 
   exists = FALSE;
-  children = gtk_container_get_children (GTK_CONTAINER (self->tags_list_box));
-  for (l = children; l != NULL; l = l->next) {
-    title = g_object_get_data (G_OBJECT (l->data), "title");
-    type = g_object_get_data (G_OBJECT (l->data), "type");
+
+  while ((row = gtk_list_box_get_row_at_index (GTK_LIST_BOX (self->tags_list_box), i++))) {
+    const char *title = g_object_get_data (G_OBJECT (row), "title");
+    const char *type = g_object_get_data (G_OBJECT (row), "type");
 
     if (g_strcmp0 (title, tag) == 0 &&
         g_strcmp0 (type, EPHY_LIST_BOX_ROW_TYPE_TAG) == 0) {
@@ -119,11 +130,10 @@ ephy_bookmarks_popover_bookmark_tag_added_cb (EphyBookmarksPopover *self,
       break;
     }
   }
-  g_list_free (children);
 
   if (!exists) {
-    tag_row = create_tag_row (tag);
-    gtk_container_add (GTK_CONTAINER (self->tags_list_box), tag_row);
+    GtkWidget *tag_row = create_tag_row (tag);
+    gtk_list_box_append (GTK_LIST_BOX (self->tags_list_box), tag_row);
   }
 }
 
@@ -133,12 +143,8 @@ ephy_bookmarks_popover_bookmark_tag_removed_cb (EphyBookmarksPopover *self,
                                                 const char           *tag,
                                                 EphyBookmarksManager *manager)
 {
-  GtkWidget *row;
-  GList *children;
-  GList *l;
   const char *visible_stack_child;
   gboolean exists;
-  gpointer title;
 
   g_assert (EPHY_IS_BOOKMARK (bookmark));
   g_assert (EPHY_IS_BOOKMARKS_POPOVER (self));
@@ -146,13 +152,15 @@ ephy_bookmarks_popover_bookmark_tag_removed_cb (EphyBookmarksPopover *self,
   /* If the bookmark has 0 tags after removing one, we add it to the tags list
    * box */
   if (g_sequence_is_empty (ephy_bookmark_get_tags (bookmark))) {
+    GtkListBoxRow *row;
+    int i = 0;
+
     exists = FALSE;
-    children = gtk_container_get_children (GTK_CONTAINER (self->tags_list_box));
-    for (l = children; l != NULL; l = l->next) {
-      const char *type = g_object_get_data (l->data, "type");
+    while ((row = gtk_list_box_get_row_at_index (GTK_LIST_BOX (self->tags_list_box), i++))) {
+      const char *type = g_object_get_data (G_OBJECT (row), "type");
 
       if (g_strcmp0 (type, EPHY_LIST_BOX_ROW_TYPE_BOOKMARK) == 0) {
-        const char *url = ephy_bookmark_row_get_bookmark_url (l->data);
+        const char *url = ephy_bookmark_row_get_bookmark_url (EPHY_BOOKMARK_ROW (row));
 
         if (g_strcmp0 (ephy_bookmark_get_url (bookmark), url) == 0) {
           exists = TRUE;
@@ -160,11 +168,10 @@ ephy_bookmarks_popover_bookmark_tag_removed_cb (EphyBookmarksPopover *self,
         }
       }
     }
-    g_list_free (children);
 
     if (!exists) {
-      row = create_bookmark_row (bookmark, self);
-      gtk_container_add (GTK_CONTAINER (self->tags_list_box), row);
+      GtkWidget *row = create_bookmark_row (bookmark, self);
+      gtk_list_box_append (GTK_LIST_BOX (self->tags_list_box), row);
     }
   }
 
@@ -173,29 +180,25 @@ ephy_bookmarks_popover_bookmark_tag_removed_cb (EphyBookmarksPopover *self,
   visible_stack_child = gtk_stack_get_visible_child_name (GTK_STACK (self->toplevel_stack));
   if (g_strcmp0 (visible_stack_child, "tag_detail") == 0 &&
       g_strcmp0 (self->tag_detail_tag, tag) == 0) {
-    remove_bookmark_row_from_container (GTK_CONTAINER (self->tag_detail_list_box),
-                                        ephy_bookmark_get_url (bookmark));
+    remove_bookmark_row (GTK_LIST_BOX (self->tag_detail_list_box),
+                         ephy_bookmark_get_url (bookmark));
 
     /* If we removed the tag's last bookmark, switch back to the tags list. */
-    if (ephy_bookmarks_manager_has_bookmarks_with_tag (self->manager, tag)) {
-      GActionGroup *group;
-      GAction *action;
-
-      group = gtk_widget_get_action_group (GTK_WIDGET (self), "popover");
-      action = g_action_map_lookup_action (G_ACTION_MAP (group), "tag-detail-back");
-      g_action_activate (action, NULL);
-    }
+    if (ephy_bookmarks_manager_has_bookmarks_with_tag (self->manager, tag))
+      tag_detail_back (self);
   }
 
   /* If the tag no longer contains bookmarks, remove it from the tags list */
   if (ephy_bookmarks_manager_has_bookmarks_with_tag (self->manager, tag)) {
-    children = gtk_container_get_children (GTK_CONTAINER (self->tags_list_box));
-    for (l = children; l != NULL; l = l->next) {
-      title = g_object_get_data (G_OBJECT (l->data), "title");
+    GtkListBoxRow *row;
+    int i = 0;
+
+    while ((row = gtk_list_box_get_row_at_index (GTK_LIST_BOX (self->tags_list_box), i++))) {
+      const char *title = g_object_get_data (G_OBJECT (row), "title");
+
       if (g_strcmp0 (title, tag) == 0)
-        gtk_container_remove (GTK_CONTAINER (self->tags_list_box), GTK_WIDGET (l->data));
+        gtk_list_box_remove (GTK_LIST_BOX (self->tags_list_box), GTK_WIDGET (row));
     }
-    g_list_free (children);
   }
 }
 
@@ -237,9 +240,9 @@ create_tag_row (const char *tag)
   box = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
 
   if (g_strcmp0 (tag, EPHY_BOOKMARKS_FAVORITES_TAG) == 0) {
-    image = gtk_image_new_from_icon_name ("emblem-favorite-symbolic", GTK_ICON_SIZE_MENU);
+    image = gtk_image_new_from_icon_name ("emblem-favorite-symbolic");
   } else {
-    image = gtk_image_new_from_icon_name ("ephy-bookmark-tag-symbolic", GTK_ICON_SIZE_MENU);
+    image = gtk_image_new_from_icon_name ("ephy-bookmark-tag-symbolic");
   }
   label = gtk_label_new (tag);
 
@@ -247,11 +250,10 @@ create_tag_row (const char *tag)
   gtk_label_set_xalign (GTK_LABEL (label), 0);
   gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
 
-  gtk_box_pack_start (GTK_BOX (box), image, FALSE, TRUE, 0);
-  gtk_box_pack_start (GTK_BOX (box), label, FALSE, TRUE, 0);
+  gtk_box_append (GTK_BOX (box), image);
+  gtk_box_append (GTK_BOX (box), label);
 
-  gtk_container_add (GTK_CONTAINER (row), box);
-  gtk_widget_show_all (row);
+  gtk_list_box_row_set_child (GTK_LIST_BOX_ROW (row), box);
 
   return row;
 }
@@ -269,7 +271,7 @@ ephy_bookmarks_popover_bookmark_added_cb (EphyBookmarksPopover *self,
 
   if (g_sequence_is_empty (ephy_bookmark_get_tags (bookmark))) {
     row = create_bookmark_row (bookmark, self);
-    gtk_container_add (GTK_CONTAINER (self->tags_list_box), row);
+    gtk_list_box_append (GTK_LIST_BOX (self->tags_list_box), row);
   }
 
   if (strcmp (gtk_stack_get_visible_child_name (GTK_STACK (self->toplevel_stack)), "empty-state") == 0)
@@ -285,22 +287,17 @@ ephy_bookmarks_popover_bookmark_removed_cb (EphyBookmarksPopover *self,
   g_assert (EPHY_IS_BOOKMARK (bookmark));
   g_assert (EPHY_IS_BOOKMARKS_MANAGER (manager));
 
-  remove_bookmark_row_from_container (GTK_CONTAINER (self->tags_list_box),
-                                      ephy_bookmark_get_url (bookmark));
-  remove_bookmark_row_from_container (GTK_CONTAINER (self->tag_detail_list_box),
-                                      ephy_bookmark_get_url (bookmark));
+  remove_bookmark_row (GTK_LIST_BOX (self->tags_list_box),
+                       ephy_bookmark_get_url (bookmark));
+  remove_bookmark_row (GTK_LIST_BOX (self->tag_detail_list_box),
+                       ephy_bookmark_get_url (bookmark));
 
   if (g_list_model_get_n_items (G_LIST_MODEL (self->manager)) == 0) {
     gtk_stack_set_visible_child_name (GTK_STACK (self->toplevel_stack), "empty-state");
   } else if (g_strcmp0 (gtk_stack_get_visible_child_name (GTK_STACK (self->toplevel_stack)), "tag_detail") == 0 &&
              ephy_bookmarks_manager_has_bookmarks_with_tag (self->manager, self->tag_detail_tag)) {
     /* If we removed the tag's last bookmark, switch back to the tags list. */
-    GActionGroup *group;
-    GAction *action;
-
-    group = gtk_widget_get_action_group (GTK_WIDGET (self), "popover");
-    action = g_action_map_lookup_action (G_ACTION_MAP (group), "tag-detail-back");
-    g_action_activate (action, NULL);
+    tag_detail_back (self);
   }
 }
 
@@ -316,7 +313,7 @@ ephy_bookmarks_popover_tag_created_cb (EphyBookmarksPopover *self,
   g_assert (EPHY_IS_BOOKMARKS_MANAGER (manager));
 
   tag_row = create_tag_row (tag);
-  gtk_container_add (GTK_CONTAINER (self->tags_list_box), tag_row);
+  gtk_list_box_append (GTK_LIST_BOX (self->tags_list_box), tag_row);
 }
 
 static void
@@ -331,17 +328,12 @@ ephy_bookmarks_popover_tag_deleted_cb (EphyBookmarksPopover *self,
   g_assert (EPHY_IS_BOOKMARKS_MANAGER (manager));
 
   row = gtk_list_box_get_row_at_index (GTK_LIST_BOX (self->tags_list_box), position);
-  gtk_container_remove (GTK_CONTAINER (self->tags_list_box),
-                        GTK_WIDGET (row));
+  gtk_list_box_remove (GTK_LIST_BOX (self->tags_list_box),
+                       GTK_WIDGET (row));
 
   if (g_strcmp0 (gtk_stack_get_visible_child_name (GTK_STACK (self->toplevel_stack)), "tag_detail") == 0 &&
       g_strcmp0 (self->tag_detail_tag, tag) == 0) {
-    GActionGroup *group;
-    GAction *action;
-
-    group = gtk_widget_get_action_group (GTK_WIDGET (self), "popover");
-    action = g_action_map_lookup_action (G_ACTION_MAP (group), "tag-detail-back");
-    g_action_activate (action, NULL);
+    tag_detail_back (self);
   }
 }
 
@@ -375,25 +367,6 @@ tags_list_box_sort_func (GtkListBoxRow *row1,
   return g_strcmp0 (title1, title2);
 }
 
-static void
-ephy_bookmarks_popover_actions_tag_detail_back (GSimpleAction *action,
-                                                GVariant      *value,
-                                                gpointer       user_data)
-{
-  EphyBookmarksPopover *self = user_data;
-  GList *children;
-
-  g_assert (EPHY_IS_BOOKMARKS_POPOVER (self));
-
-  gtk_stack_set_visible_child_name (GTK_STACK (self->toplevel_stack),
-                                    "default");
-
-  children = gtk_container_get_children (GTK_CONTAINER (self->tag_detail_list_box));
-  for (GList *l = children; l != NULL; l = l->next)
-    gtk_container_remove (GTK_CONTAINER (self->tag_detail_list_box), l->data);
-  g_list_free (children);
-}
-
 static void
 ephy_bookmarks_popover_show_tag_detail (EphyBookmarksPopover *self,
                                         const char           *tag)
@@ -409,7 +382,7 @@ ephy_bookmarks_popover_show_tag_detail (EphyBookmarksPopover *self,
     GtkWidget *row;
 
     row = create_bookmark_row (bookmark, self);
-    gtk_container_add (GTK_CONTAINER (self->tag_detail_list_box), row);
+    gtk_list_box_append (GTK_LIST_BOX (self->tag_detail_list_box), row);
   }
 
   gtk_label_set_label (GTK_LABEL (self->tag_detail_label), tag);
@@ -424,62 +397,52 @@ ephy_bookmarks_popover_show_tag_detail (EphyBookmarksPopover *self,
 }
 
 static void
-ephy_bookmarks_popover_open_bookmark (EphyBookmarksPopover *self,
-                                      GtkListBoxRow        *row)
-{
-  GtkWidget *window;
-  GActionGroup *action_group;
-  GAction *action;
-  const char *url;
-
-  window = gtk_widget_get_ancestor (GTK_WIDGET (self), EPHY_TYPE_WINDOW);
-  g_assert (EPHY_IS_WINDOW (window));
-  action_group = gtk_widget_get_action_group (window, "win");
-  g_assert (action_group != NULL);
-  action = g_action_map_lookup_action (G_ACTION_MAP (action_group), "open-bookmark");
-  g_assert (action != NULL);
-  url = ephy_bookmark_row_get_bookmark_url (EPHY_BOOKMARK_ROW (row));
-
-  g_action_activate (action, g_variant_new_string (url));
-}
-
-static void
-ephy_bookmarks_popover_list_box_row_activated_cb (EphyBookmarksPopover *self,
-                                                  GtkListBoxRow        *row,
-                                                  GtkListBox           *box)
+row_clicked_cb (GtkGesture           *gesture,
+                int                   n_click,
+                double                x,
+                double                y,
+                EphyBookmarksPopover *self)
 {
+  GtkWidget *list;
+  GtkListBoxRow *row;
+  guint button;
   const char *type;
-  const char *tag;
 
-  g_assert (EPHY_IS_BOOKMARKS_POPOVER (self));
-  g_assert (GTK_IS_LIST_BOX_ROW (row));
-  g_assert (GTK_IS_LIST_BOX (box));
+  button = gtk_gesture_single_get_current_button (GTK_GESTURE_SINGLE (gesture));
+
+  if (button != GDK_BUTTON_PRIMARY && button != GDK_BUTTON_MIDDLE) {
+    gtk_gesture_set_state (gesture, GTK_EVENT_SEQUENCE_DENIED);
+    return;
+  }
+
+  list = gtk_event_controller_get_widget (GTK_EVENT_CONTROLLER (gesture));
+
+  g_assert (GTK_IS_LIST_BOX (list));
+
+  row = gtk_list_box_get_row_at_y (GTK_LIST_BOX (list), y);
+
+  gtk_gesture_set_state (gesture, GTK_EVENT_SEQUENCE_CLAIMED);
 
   type = g_object_get_data (G_OBJECT (row), "type");
   if (g_strcmp0 (type, EPHY_LIST_BOX_ROW_TYPE_BOOKMARK) == 0) {
-    ephy_bookmarks_popover_open_bookmark (self, row);
+    GdkModifierType modifiers;
+    EphyLinkFlags flags;
+
+    modifiers = gtk_event_controller_get_current_event_state (GTK_EVENT_CONTROLLER (gesture));
+    modifiers &= gtk_accelerator_get_default_mod_mask ();
+
+    flags = ephy_link_flags_from_modifiers (modifiers, button == GDK_BUTTON_MIDDLE);
+
+    ephy_bookmark_row_open (EPHY_BOOKMARK_ROW (row), flags);
+
+    if (button != GDK_BUTTON_MIDDLE)
+      gtk_popover_popdown (GTK_POPOVER (self));
   } else {
-    tag = g_object_get_data (G_OBJECT (row), "title");
+    const char *tag = g_object_get_data (G_OBJECT (row), "title");
     ephy_bookmarks_popover_show_tag_detail (self, tag);
   }
 }
 
-static gboolean
-ephy_bookmarks_popover_list_box_button_release_event_cb (EphyBookmarksPopover *self,
-                                                         GdkEvent             *event,
-                                                         GtkListBox           *box)
-{
-  GdkEventButton *event_button = (GdkEventButton *)event;
-  GtkListBoxRow *row = gtk_list_box_get_row_at_y (box, event_button->y);
-
-  if (event_button->button != GDK_BUTTON_MIDDLE)
-    return GDK_EVENT_PROPAGATE;
-
-  ephy_bookmarks_popover_open_bookmark (self, row);
-
-  return GDK_EVENT_STOP;
-}
-
 static void
 ephy_bookmarks_popover_finalize (GObject *object)
 {
@@ -504,11 +467,10 @@ ephy_bookmarks_popover_class_init (EphyBookmarksPopoverClass *klass)
   gtk_widget_class_bind_template_child (widget_class, EphyBookmarksPopover, tags_list_box);
   gtk_widget_class_bind_template_child (widget_class, EphyBookmarksPopover, tag_detail_list_box);
   gtk_widget_class_bind_template_child (widget_class, EphyBookmarksPopover, tag_detail_label);
-}
 
-static const GActionEntry entries[] = {
-  { "tag-detail-back", ephy_bookmarks_popover_actions_tag_detail_back }
-};
+  gtk_widget_class_install_action (widget_class, "popover.tag-detail-back", NULL,
+                                   (GtkWidgetActionActivateFunc)tag_detail_back);
+}
 
 static void
 ephy_bookmarks_popover_init (EphyBookmarksPopover *self)
@@ -516,18 +478,12 @@ ephy_bookmarks_popover_init (EphyBookmarksPopover *self)
   GSequence *tags;
   GSequenceIter *iter;
   g_autoptr (GSequence) bookmarks = NULL;
-  g_autoptr (GSimpleActionGroup) group = NULL;
+  GtkGesture *gesture;
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
   self->manager = ephy_shell_get_bookmarks_manager (ephy_shell_get_default ());
 
-  group = g_simple_action_group_new ();
-  g_action_map_add_action_entries (G_ACTION_MAP (group), entries,
-                                   G_N_ELEMENTS (entries), self);
-  gtk_widget_insert_action_group (GTK_WIDGET (self), "popover",
-                                  G_ACTION_GROUP (group));
-
   gtk_list_box_bind_model (GTK_LIST_BOX (self->bookmarks_list_box),
                            G_LIST_MODEL (self->manager),
                            create_bookmark_row,
@@ -552,7 +508,7 @@ ephy_bookmarks_popover_init (EphyBookmarksPopover *self)
 
     if (!ephy_bookmarks_manager_has_bookmarks_with_tag (self->manager, tag)) {
       tag_row = create_tag_row (tag);
-      gtk_container_add (GTK_CONTAINER (self->tags_list_box), tag_row);
+      gtk_list_box_append (GTK_LIST_BOX (self->tags_list_box), tag_row);
     }
   }
 
@@ -564,8 +520,7 @@ ephy_bookmarks_popover_init (EphyBookmarksPopover *self)
     GtkWidget *bookmark_row;
 
     bookmark_row = create_bookmark_row (bookmark, self);
-    gtk_widget_show_all (bookmark_row);
-    gtk_container_add (GTK_CONTAINER (self->tags_list_box), bookmark_row);
+    gtk_list_box_append (GTK_LIST_BOX (self->tags_list_box), bookmark_row);
   }
 
   g_signal_connect_object (self->manager, "bookmark-added",
@@ -587,18 +542,20 @@ ephy_bookmarks_popover_init (EphyBookmarksPopover *self)
                            G_CALLBACK (ephy_bookmarks_popover_bookmark_tag_removed_cb),
                            self, G_CONNECT_SWAPPED);
 
-  g_signal_connect_object (self->bookmarks_list_box, "row-activated",
-                           G_CALLBACK (ephy_bookmarks_popover_list_box_row_activated_cb),
-                           self, G_CONNECT_SWAPPED);
-  g_signal_connect_object (self->tags_list_box, "row-activated",
-                           G_CALLBACK (ephy_bookmarks_popover_list_box_row_activated_cb),
-                           self, G_CONNECT_SWAPPED);
-  g_signal_connect_object (self->tag_detail_list_box, "row-activated",
-                           G_CALLBACK (ephy_bookmarks_popover_list_box_row_activated_cb),
-                           self, G_CONNECT_SWAPPED);
-  g_signal_connect_object (self->bookmarks_list_box, "button-release-event",
-                           G_CALLBACK (ephy_bookmarks_popover_list_box_button_release_event_cb),
-                           self, G_CONNECT_SWAPPED);
+  gesture = gtk_gesture_click_new ();
+  gtk_gesture_single_set_button (GTK_GESTURE_SINGLE (gesture), 0);
+  g_signal_connect (gesture, "released", G_CALLBACK (row_clicked_cb), self);
+  gtk_widget_add_controller (self->bookmarks_list_box, GTK_EVENT_CONTROLLER (gesture));
+
+  gesture = gtk_gesture_click_new ();
+  gtk_gesture_single_set_button (GTK_GESTURE_SINGLE (gesture), 0);
+  g_signal_connect (gesture, "released", G_CALLBACK (row_clicked_cb), self);
+  gtk_widget_add_controller (self->tags_list_box, GTK_EVENT_CONTROLLER (gesture));
+
+  gesture = gtk_gesture_click_new ();
+  gtk_gesture_single_set_button (GTK_GESTURE_SINGLE (gesture), 0);
+  g_signal_connect (gesture, "released", G_CALLBACK (row_clicked_cb), self);
+  gtk_widget_add_controller (self->tag_detail_list_box, GTK_EVENT_CONTROLLER (gesture));
 }
 
 EphyBookmarksPopover *
