@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
 /*
- *  Copyright © 2019-2020 Jan-Michael Brummer <jan.brummer@tabos.org>
+ *  Copyright © 2019-2022 Jan-Michael Brummer <jan.brummer@tabos.org>
  *
  *  This file is part of Epiphany.
  *
@@ -20,6 +20,7 @@
 
 #include "config.h"
 
+#include "ephy-browser-action.h"
 #include "ephy-debug.h"
 #include "ephy-embed-shell.h"
 #include "ephy-embed-prefs.h"
@@ -36,6 +37,7 @@
 #include "ephy-web-view.h"
 
 #include "api/alarms.h"
+#include "api/browseraction.h"
 #include "api/commands.h"
 #include "api/cookies.h"
 #include "api/downloads.h"
@@ -47,6 +49,7 @@
 #include "api/tabs.h"
 #include "api/windows.h"
 
+#include <adwaita.h>
 #include <json-glib/json-glib.h>
 
 static void handle_message_reply (EphyWebExtension *web_extension,
@@ -58,7 +61,9 @@ struct _EphyWebExtensionManager {
   GCancellable *cancellable;
   GPtrArray *web_extensions;
   GHashTable *page_action_map;
+
   GHashTable *browser_action_map;
+  GListStore *browser_actions;
 
   GHashTable *user_agent_overrides;
 
@@ -68,10 +73,11 @@ struct _EphyWebExtensionManager {
   GHashTable *pending_messages;
 };
 
-G_DEFINE_TYPE (EphyWebExtensionManager, ephy_web_extension_manager, G_TYPE_OBJECT)
+G_DEFINE_FINAL_TYPE (EphyWebExtensionManager, ephy_web_extension_manager, G_TYPE_OBJECT)
 
 EphyWebExtensionApiHandler api_handlers[] = {
   {"alarms", ephy_web_extension_api_alarms_handler},
+  {"browserAction", ephy_web_extension_api_browseraction_handler},
   {"commands", ephy_web_extension_api_commands_handler},
   {"cookies", ephy_web_extension_api_cookies_handler},
   {"downloads", ephy_web_extension_api_downloads_handler},
@@ -87,6 +93,7 @@ EphyWebExtensionApiHandler api_handlers[] = {
 
 enum {
   CHANGED,
+  SHOW_BROWSER_ACTION,
   LAST_SIGNAL
 };
 
@@ -144,7 +151,7 @@ on_web_extension_loaded (GObject      *source_object,
 {
   GFile *target = G_FILE (source_object);
   g_autoptr (GError) error = NULL;
-  EphyWebExtension *web_extension;
+  g_autoptr (EphyWebExtension) web_extension = NULL;
   EphyWebExtensionManager *self = EPHY_WEB_EXTENSION_MANAGER (user_data);
 
   web_extension = ephy_web_extension_load_finished (source_object, result, &error);
@@ -154,7 +161,6 @@ on_web_extension_loaded (GObject      *source_object,
   }
 
   ephy_web_extension_manager_add_to_list (self, web_extension);
-  g_object_unref (web_extension);
 
   if (ephy_web_extension_manager_is_active (self, web_extension))
     ephy_web_extension_manager_set_active (self, web_extension, TRUE);
@@ -172,7 +178,8 @@ scan_directory_ready_cb (GFile        *file,
   enumerator = g_file_enumerate_children_finish (file, result, &error);
 
   if (error) {
-    g_warning ("Failed to scan extensions directory: %s", error->message);
+    if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND))
+      g_warning ("Failed to scan extensions directory: %s", error->message);
     return;
   }
 
@@ -206,12 +213,6 @@ ephy_web_extension_manager_scan_directory_async (EphyWebExtensionManager *self,
                                    self);
 }
 
-static void
-destroy_widget_list (GSList *widget_list)
-{
-  g_slist_free_full (widget_list, (GDestroyNotify)gtk_widget_destroy);
-}
-
 static void
 ephy_webextension_scheme_cb (WebKitURISchemeRequest *request,
                              gpointer                user_data)
@@ -226,7 +227,7 @@ ephy_webextension_scheme_cb (WebKitURISchemeRequest *request,
   gsize length;
 
   uri = g_uri_parse (webkit_uri_scheme_request_get_uri (request),
-                     G_URI_FLAGS_ENCODED_PATH | G_URI_FLAGS_SCHEME_NORMALIZE,
+                     G_URI_FLAGS_PARSE_RELAXED | G_URI_FLAGS_ENCODED_PATH | G_URI_FLAGS_SCHEME_NORMALIZE,
                      &error);
   if (!uri) {
     webkit_uri_scheme_request_finish_error (request, g_steal_pointer (&error));
@@ -266,10 +267,11 @@ ephy_web_extension_manager_constructed (GObject *object)
   EphyWebExtensionManager *self = EPHY_WEB_EXTENSION_MANAGER (object);
   g_autofree char *dir = g_build_filename (ephy_default_profile_dir (), "web_extensions", NULL);
 
-  self->background_web_views = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, (GDestroyNotify)gtk_widget_destroy);
+  self->background_web_views = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, (GDestroyNotify)g_object_unref);
   self->popup_web_views = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, (GDestroyNotify)g_ptr_array_free);
   self->page_action_map = g_hash_table_new_full (NULL, NULL, NULL, (GDestroyNotify)g_hash_table_destroy);
-  self->browser_action_map = g_hash_table_new_full (NULL, NULL, NULL, (GDestroyNotify)destroy_widget_list);
+  self->browser_action_map = g_hash_table_new_full (NULL, NULL, NULL, (GDestroyNotify)g_object_unref);
+  self->browser_actions = g_list_store_new (EPHY_TYPE_BROWSER_ACTION);
   self->pending_messages = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, (GDestroyNotify)g_hash_table_destroy);
   self->web_extensions = g_ptr_array_new_full (0, g_object_unref);
   self->user_agent_overrides = create_user_agent_overrides ();
@@ -284,8 +286,12 @@ ephy_web_extension_manager_dispose (GObject *object)
 
   ephy_web_extension_api_downloads_dispose (self);
 
+  g_list_store_remove_all (self->browser_actions);
+
   g_clear_pointer (&self->background_web_views, g_hash_table_destroy);
   g_clear_pointer (&self->popup_web_views, g_hash_table_destroy);
+  g_clear_object (&self->browser_actions);
+  g_clear_pointer (&self->browser_action_map, g_hash_table_destroy);
   g_clear_pointer (&self->page_action_map, g_hash_table_destroy);
   g_clear_pointer (&self->pending_messages, g_hash_table_destroy);
   g_clear_pointer (&self->web_extensions, g_ptr_array_unref);
@@ -306,6 +312,14 @@ ephy_web_extension_manager_class_init (EphyWebExtensionManagerClass *klass)
                   G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
                   0, NULL, NULL, NULL,
                   G_TYPE_NONE, 0);
+
+  signals[SHOW_BROWSER_ACTION] =
+    g_signal_new ("show-browser-action",
+                  G_OBJECT_CLASS_TYPE (object_class),
+                  G_SIGNAL_RUN_FIRST,
+                  0, NULL, NULL, NULL,
+                  G_TYPE_NONE,
+                  1, EPHY_TYPE_BROWSER_ACTION);
 }
 
 static void
@@ -392,9 +406,12 @@ ephy_web_extension_manager_install (EphyWebExtensionManager *self,
   gboolean is_xpi = FALSE;
   g_autoptr (GError) error = NULL;
   g_autoptr (GFile) web_extensions_dir = NULL;
+  const char *path;
 
   web_extensions_dir = g_file_new_build_filename (ephy_default_profile_dir (), "web_extensions", NULL);
-  is_xpi = g_str_has_suffix (g_file_peek_path (file), ".xpi");
+  path = g_file_peek_path (file);
+  g_assert (path);
+  is_xpi = g_str_has_suffix (path, ".xpi");
 
   /* FIXME: Make this async. */
 
@@ -519,10 +536,9 @@ ephy_web_extension_manager_append_context_menu (EphyWebExtensionManager *self,
   }
 }
 
-static gboolean
-page_action_clicked (GtkWidget      *event_box,
-                     GdkEventButton *event,
-                     gpointer        user_data)
+static void
+page_action_clicked (GtkButton *button,
+                     gpointer   user_data)
 {
   EphyWebExtension *web_extension = EPHY_WEB_EXTENSION (user_data);
   EphyShell *shell = ephy_shell_get_default ();
@@ -546,13 +562,13 @@ page_action_clicked (GtkWidget      *event_box,
   json = json_to_string (root, FALSE);
 
   script = g_strdup_printf ("window.browser.pageAction.onClicked._emit(%s);", json);
-  webkit_web_view_run_javascript (web_view,
-                                  script,
-                                  NULL,
-                                  NULL,
-                                  NULL);
-
-  return GDK_EVENT_STOP;
+  webkit_web_view_evaluate_javascript (web_view,
+                                       script, -1,
+                                       NULL,
+                                       NULL,
+                                       NULL,
+                                       NULL,
+                                       NULL);
 }
 
 static GtkWidget *
@@ -560,20 +576,23 @@ create_page_action_widget (EphyWebExtensionManager *self,
                            EphyWebExtension        *web_extension)
 {
   GtkWidget *image;
-  GtkWidget *event_box;
-  GtkStyleContext *context;
+  GtkWidget *button;
+
+  button = gtk_button_new ();
+  gtk_widget_set_valign (button, GTK_ALIGN_CENTER);
 
-  /* Create new event box with page action */
-  event_box = gtk_event_box_new ();
   image = gtk_image_new ();
-  gtk_container_add (GTK_CONTAINER (event_box), image);
-  g_signal_connect_object (event_box, "button_press_event", G_CALLBACK (page_action_clicked), web_extension, 0);
-  gtk_widget_show_all (event_box);
+  gtk_image_set_pixel_size (GTK_IMAGE (image), 16);
+  gtk_button_set_child (GTK_BUTTON (button), image);
 
-  context = gtk_widget_get_style_context (image);
-  gtk_style_context_add_class (context, "entry_icon");
+  gtk_widget_add_css_class (button, "image-button");
+  gtk_widget_add_css_class (button, "entry-icon");
+  gtk_widget_add_css_class (button, "end");
 
-  return g_object_ref (event_box);
+  g_signal_connect_object (button, "clicked",
+                           G_CALLBACK (page_action_clicked), web_extension, 0);
+
+  return g_object_ref (button);
 }
 
 static void
@@ -856,11 +875,22 @@ send_to_page_ready_cb (WebKitWebView    *web_view,
                        EphyWebExtension *web_extension)
 {
   g_autoptr (WebKitUserMessage) response = webkit_web_view_send_message_to_page_finish (web_view, result, NULL);
+  (void)response;
 
   update_translations (web_extension);
   add_content_scripts (web_extension, EPHY_WEB_VIEW (web_view));
 }
 
+static void
+destroy_page_action (GtkWidget *action)
+{
+  EphyLocationEntry *entry;
+
+  entry = EPHY_LOCATION_ENTRY (gtk_widget_get_ancestor (action, EPHY_TYPE_LOCATION_ENTRY));
+
+  ephy_location_entry_page_action_remove (entry, action);
+}
+
 static void
 ephy_web_extension_manager_add_web_extension_to_webview (EphyWebExtensionManager *self,
                                                          EphyWebExtension        *web_extension,
@@ -879,7 +909,7 @@ ephy_web_extension_manager_add_web_extension_to_webview (EphyWebExtensionManager
 
       table = g_hash_table_lookup (self->page_action_map, web_extension);
       if (!table) {
-        table = g_hash_table_new_full (NULL, NULL, NULL, (GDestroyNotify)gtk_widget_destroy);
+        table = g_hash_table_new_full (NULL, NULL, NULL, (GDestroyNotify)destroy_page_action);
         g_hash_table_insert (self->page_action_map, web_extension, table);
       }
 
@@ -897,15 +927,15 @@ ephy_web_extension_manager_add_web_extension_to_webview (EphyWebExtensionManager
 }
 
 static void
-page_attached_cb (HdyTabView *tab_view,
-                  HdyTabPage *page,
+page_attached_cb (AdwTabView *tab_view,
+                  AdwTabPage *page,
                   gint        position,
                   gpointer    user_data)
 {
   EphyWebExtension *web_extension = EPHY_WEB_EXTENSION (user_data);
-  GtkWidget *child = hdy_tab_page_get_child (page);
+  GtkWidget *child = adw_tab_page_get_child (page);
   EphyWebView *web_view = ephy_embed_get_web_view (EPHY_EMBED (child));
-  EphyWindow *window = EPHY_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (tab_view)));
+  EphyWindow *window = EPHY_WINDOW (gtk_widget_get_root (GTK_WIDGET (tab_view)));
   EphyWebExtensionManager *self = ephy_web_extension_manager_get_default ();
 
   ephy_web_extension_manager_add_web_extension_to_webview (self, web_extension, window, web_view);
@@ -920,9 +950,9 @@ init_web_extension_api (WebKitWebContext *web_context,
   g_autofree char *translations = get_translation_contents (web_extension);
 
 #if DEVELOPER_MODE
-  webkit_web_context_set_web_extensions_directory (web_context, BUILD_ROOT "/embed/web-process-extension");
+  webkit_web_context_set_web_process_extensions_directory (web_context, BUILD_ROOT "/embed/web-process-extension");
 #else
-  webkit_web_context_set_web_extensions_directory (web_context, EPHY_WEB_PROCESS_EXTENSIONS_DIR);
+  webkit_web_context_set_web_process_extensions_directory (web_context, EPHY_WEB_PROCESS_EXTENSIONS_DIR);
 #endif
 
   user_data = g_variant_new ("(smsbbbs)",
@@ -932,7 +962,7 @@ init_web_extension_api (WebKitWebContext *web_context,
                              FALSE /* private_profile */,
                              TRUE /* is_webextension */,
                              translations);
-  webkit_web_context_set_web_extensions_initialization_user_data (web_context, g_steal_pointer (&user_data));
+  webkit_web_context_set_web_process_extensions_initialization_user_data (web_context, g_steal_pointer (&user_data));
 }
 
 static gboolean
@@ -984,10 +1014,10 @@ GtkWidget *
 ephy_web_extensions_manager_create_web_extensions_webview (EphyWebExtension *web_extension)
 {
   EphyWebExtensionManager *manager = ephy_web_extension_manager_get_default ();
-  g_autoptr (WebKitWebsiteDataManager) data_manager = NULL;
   g_autoptr (WebKitSettings) settings = NULL;
-  WebKitWebContext *web_context;
+  WebKitWebContext *web_context = NULL;
   GtkWidget *web_view;
+  WebKitWebView *background_view;
   const char *custom_user_agent;
 
   settings = webkit_settings_new_with_settings ("enable-write-console-messages-to-stdout", TRUE,
@@ -1003,19 +1033,20 @@ ephy_web_extensions_manager_create_web_extensions_webview (EphyWebExtension *web
   else
     webkit_settings_set_user_agent_with_application_details (settings, "Epiphany", EPHY_VERSION);
 
-  data_manager = webkit_web_context_get_website_data_manager (ephy_embed_shell_get_web_context (ephy_embed_shell_get_default ()));
-  web_context = webkit_web_context_new_with_website_data_manager (data_manager);
-  webkit_web_context_set_sandbox_enabled (web_context, TRUE);
-
-  webkit_web_context_register_uri_scheme (web_context, "ephy-webextension", ephy_webextension_scheme_cb, web_extension, NULL);
-  webkit_security_manager_register_uri_scheme_as_secure (webkit_web_context_get_security_manager (web_context),
-                                                         "ephy-webextension");
-  g_signal_connect_object (web_context, "initialize-web-extensions", G_CALLBACK (init_web_extension_api), web_extension, 0);
+  /* If there is a background view the web context is shared with the related-view property. */
+  background_view = ephy_web_extension_manager_get_background_web_view (manager, web_extension);
+  if (!background_view) {
+    web_context = webkit_web_context_new ();
+    webkit_web_context_register_uri_scheme (web_context, "ephy-webextension", ephy_webextension_scheme_cb, web_extension, NULL);
+    webkit_security_manager_register_uri_scheme_as_secure (webkit_web_context_get_security_manager (web_context),
+                                                           "ephy-webextension");
+    g_signal_connect_object (web_context, "initialize-web-process-extensions", G_CALLBACK (init_web_extension_api), web_extension, 0);
+  }
 
   web_view = g_object_new (WEBKIT_TYPE_WEB_VIEW,
                            "web-context", web_context,
                            "settings", settings,
-                           "related-view", ephy_web_extension_manager_get_background_web_view (manager, web_extension),
+                           "related-view", background_view,
                            "default-content-security-policy", ephy_web_extension_get_content_security_policy (web_extension),
                            "web-extension-mode", WEBKIT_WEB_EXTENSION_MODE_MANIFESTV2,
                            NULL);
@@ -1036,7 +1067,7 @@ on_popup_load_changed (WebKitWebView   *web_view,
   /* TODO: Other browsers resize on DOM changes.
    *       We also need to limit this to a max size of 800x600 like Firefox. */
   if (load_event == WEBKIT_LOAD_FINISHED)
-    gtk_widget_show (GTK_WIDGET (web_view));
+    gtk_widget_set_visible (GTK_WIDGET (web_view), TRUE);
 }
 
 static void
@@ -1066,17 +1097,20 @@ ephy_web_extension_manager_register_popup_view (EphyWebExtensionManager *manager
   g_signal_connect (web_view, "destroy", G_CALLBACK (on_popup_view_destroyed), web_extension);
 }
 
-static GtkWidget *
-create_browser_popup (EphyWebExtension *web_extension)
+GtkWidget *
+ephy_web_extension_manager_create_browser_popup (EphyWebExtensionManager *self,
+                                                 EphyWebExtension        *web_extension)
 {
-  EphyWebExtensionManager *manager = ephy_web_extension_manager_get_default ();
   GtkWidget *web_view;
   g_autofree char *popup_uri = NULL;
   const char *popup;
 
   web_view = ephy_web_extensions_manager_create_web_extensions_webview (web_extension);
-  gtk_widget_hide (web_view); /* Shown in on_popup_load_changed. */
-  ephy_web_extension_manager_register_popup_view (manager, web_extension, web_view);
+  gtk_widget_set_hexpand (web_view, TRUE);
+  gtk_widget_set_vexpand (web_view, TRUE);
+  gtk_widget_set_visible (web_view, FALSE); /* Shown in on_popup_load_changed. */
+
+  ephy_web_extension_manager_register_popup_view (self, web_extension, web_view);
 
   popup = ephy_web_extension_get_browser_popup (web_extension);
   popup_uri = g_strdup_printf ("ephy-webextension://%s/%s", ephy_web_extension_get_guid (web_extension), popup);
@@ -1086,72 +1120,30 @@ create_browser_popup (EphyWebExtension *web_extension)
   return web_view;
 }
 
-static gboolean
-on_browser_action_clicked (GtkWidget *event_box,
-                           gpointer   user_data)
+void
+ephy_web_extension_manager_browseraction_set_badge_text (EphyWebExtensionManager *self,
+                                                         EphyWebExtension        *web_extension,
+                                                         const char              *text)
 {
-  EphyWebExtension *web_extension = EPHY_WEB_EXTENSION (user_data);
-  EphyWebExtensionManager *self = ephy_web_extension_manager_get_default ();
-  g_autofree char *script = NULL;
-  WebKitWebView *web_view = ephy_web_extension_manager_get_background_web_view (self, web_extension);
+  EphyBrowserAction *action = EPHY_BROWSER_ACTION (g_hash_table_lookup (self->browser_action_map, web_extension));
 
-  script = g_strdup_printf ("window.browser.browserAction.onClicked._emit();");
+  if (!action)
+    return;
 
-  webkit_web_view_run_javascript (web_view,
-                                  script,
-                                  NULL,
-                                  NULL,
-                                  NULL);
-
-  return GDK_EVENT_STOP;
+  ephy_browser_action_set_badge_text (action, text);
 }
 
-static void
-on_browser_action_visible_changed (GtkWidget  *popover,
-                                   GParamSpec *pspec,
-                                   gpointer    user_data)
+void
+ephy_web_extension_manager_browseraction_set_badge_background_color (EphyWebExtensionManager *self,
+                                                                     EphyWebExtension        *web_extension,
+                                                                     GdkRGBA                 *color)
 {
-  EphyWebExtension *web_extension = EPHY_WEB_EXTENSION (user_data);
-  GtkWidget *child;
+  EphyBrowserAction *action = EPHY_BROWSER_ACTION (g_hash_table_lookup (self->browser_action_map, web_extension));
 
-  if (gtk_widget_get_visible (popover)) {
-    child = create_browser_popup (web_extension);
-    gtk_container_add (GTK_CONTAINER (popover), child);
-  } else {
-    child = gtk_bin_get_child (GTK_BIN (popover));
-    gtk_container_remove (GTK_CONTAINER (popover), child);
-  }
-}
+  if (!action)
+    return;
 
-GtkWidget *
-create_browser_action (EphyWebExtension *web_extension)
-{
-  GtkWidget *button;
-  GtkWidget *image;
-  GtkWidget *popover;
-  GdkPixbuf *pixbuf;
-
-  pixbuf = ephy_web_extension_browser_action_get_icon (web_extension, 16);
-  if (pixbuf)
-    image = gtk_image_new_from_pixbuf (pixbuf);
-  else
-    image = gtk_image_new_from_icon_name ("application-x-addon-symbolic", GTK_ICON_SIZE_BUTTON);
-
-  if (ephy_web_extension_get_browser_popup (web_extension)) {
-    button = gtk_menu_button_new ();
-    popover = gtk_popover_new (NULL);
-    g_signal_connect (popover, "notify::visible", G_CALLBACK (on_browser_action_visible_changed), web_extension);
-    gtk_menu_button_set_popover (GTK_MENU_BUTTON (button), popover);
-    gtk_button_set_image (GTK_BUTTON (button), image);
-  } else {
-    button = gtk_button_new ();
-    g_signal_connect_object (button, "clicked", G_CALLBACK (on_browser_action_clicked), web_extension, 0);
-    gtk_button_set_image (GTK_BUTTON (button), image);
-  }
-
-  gtk_widget_set_visible (button, TRUE);
-
-  return button;
+  ephy_browser_action_set_badge_background_color (action, color);
 }
 
 void
@@ -1160,7 +1152,7 @@ ephy_web_extension_manager_add_web_extension_to_window (EphyWebExtensionManager
                                                         EphyWindow              *window)
 {
   EphyTabView *tab_view = ephy_window_get_tab_view (EPHY_WINDOW (window));
-  HdyTabView *view = ephy_tab_view_get_tab_view (tab_view);
+  AdwTabView *view = ephy_tab_view_get_tab_view (tab_view);
 
   if (!ephy_web_extension_manager_is_active (self, web_extension))
     return;
@@ -1173,16 +1165,6 @@ ephy_web_extension_manager_add_web_extension_to_window (EphyWebExtensionManager
     ephy_web_extension_manager_add_web_extension_to_webview (self, web_extension, window, web_view);
   }
 
-  if (ephy_web_extension_has_browser_action (web_extension)) {
-    GtkWidget *browser_action_widget = create_browser_action (web_extension);
-    GSList *widget_list = g_hash_table_lookup (self->browser_action_map, web_extension);
-
-    ephy_header_bar_add_browser_action (EPHY_HEADER_BAR (ephy_window_get_header_bar (window)), browser_action_widget);
-
-    g_hash_table_steal (self->browser_action_map, web_extension); /* Avoid freeing list. */
-    g_hash_table_insert (self->browser_action_map, web_extension, g_slist_append (widget_list, browser_action_widget));
-  }
-
   ephy_web_extension_manager_update_location_entry (self, window);
   g_signal_connect_object (view, "page-attached", G_CALLBACK (page_attached_cb), web_extension, 0);
 }
@@ -1233,7 +1215,7 @@ ephy_web_extension_manager_remove_web_extension_from_window (EphyWebExtensionMan
                                                              EphyWindow              *window)
 {
   EphyTabView *tab_view = ephy_window_get_tab_view (EPHY_WINDOW (window));
-  HdyTabView *view = ephy_tab_view_get_tab_view (tab_view);
+  AdwTabView *view = ephy_tab_view_get_tab_view (tab_view);
 
   if (ephy_web_extension_manager_is_active (self, web_extension))
     return;
@@ -1306,12 +1288,48 @@ typedef struct {
   guint64 window_id;
 } WindowAddedCallbackData;
 
+static void
+on_page_attached (AdwTabView *self,
+                  AdwTabPage *page,
+                  gint        position,
+                  gpointer    user_data)
+{
+  EphyWebExtensionManager *manager = ephy_web_extension_manager_get_default ();
+  EphyWebExtension *web_extension = user_data;
+  g_autoptr (JsonNode) json = NULL;
+  GtkWidget *embed = adw_tab_page_get_child (page);
+  EphyWebView *view;
+
+  view = ephy_embed_get_web_view (EPHY_EMBED (embed));
+  json = ephy_web_extension_api_tabs_create_tab_object (web_extension, view);
+  ephy_web_extension_manager_emit_in_extension_views (manager, web_extension, "tabs.onCreated", json_to_string (json, FALSE));
+}
+
+static void
+on_page_detached (AdwTabView *self,
+                  AdwTabPage *page,
+                  gint        position,
+                  gpointer    user_data)
+{
+  EphyWebExtensionManager *manager = ephy_web_extension_manager_get_default ();
+  EphyWebExtension *web_extension = user_data;
+  g_autofree char *tab_json = NULL;
+  GtkWidget *embed = adw_tab_page_get_child (page);
+  EphyWebView *view;
+
+  view = ephy_embed_get_web_view (EPHY_EMBED (embed));
+  tab_json = g_strdup_printf ("%ld", ephy_web_view_get_uid (view));
+  ephy_web_extension_manager_emit_in_extension_views (manager, web_extension, "tabs.onRemoved", tab_json);
+}
+
 static gboolean
 application_window_added_timeout_cb (gpointer user_data)
 {
   WindowAddedCallbackData *data = user_data;
   EphyWebExtensionManager *manager = ephy_web_extension_manager_get_default ();
   EphyWindow *window = ephy_web_extension_api_windows_get_window_for_id (data->window_id);
+  EphyTabView *tab_view;
+  AdwTabView *adw_tab_view;
   g_autofree char *window_json = NULL;
 
   /* It is possible the window was removed before this timeout. */
@@ -1320,6 +1338,12 @@ application_window_added_timeout_cb (gpointer user_data)
 
   window_json = ephy_web_extension_api_windows_create_window_json (data->web_extension, window);
   ephy_web_extension_manager_emit_in_extension_views (manager, data->web_extension, "windows.onCreated", window_json);
+
+  tab_view = ephy_window_get_tab_view (window);
+  adw_tab_view = ephy_tab_view_get_tab_view (tab_view);
+  g_signal_connect (G_OBJECT (adw_tab_view), "page-attached", G_CALLBACK (on_page_attached), data->web_extension);
+  g_signal_connect (G_OBJECT (adw_tab_view), "page-detached", G_CALLBACK (on_page_detached), data->web_extension);
+
   return G_SOURCE_REMOVE;
 }
 
@@ -1352,7 +1376,32 @@ application_window_removed_cb (EphyShell        *shell,
 {
   EphyWebExtensionManager *manager = ephy_web_extension_manager_get_default ();
   g_autofree char *window_json = g_strdup_printf ("%ld", ephy_window_get_uid (window));
+  EphyTabView *tab_view;
+  AdwTabView *adw_tab_view;
+
   ephy_web_extension_manager_emit_in_extension_views (manager, web_extension, "windows.onRemoved", window_json);
+
+  tab_view = ephy_window_get_tab_view (window);
+  adw_tab_view = ephy_tab_view_get_tab_view (tab_view);
+  g_signal_handlers_disconnect_by_func (G_OBJECT (adw_tab_view), G_CALLBACK (on_page_attached), web_extension);
+  g_signal_handlers_disconnect_by_func (G_OBJECT (adw_tab_view), G_CALLBACK (on_page_detached), web_extension);
+}
+
+static void
+remove_browser_action (EphyWebExtensionManager *self,
+                       EphyWebExtension        *web_extension)
+{
+  EphyBrowserAction *action;
+  guint position;
+
+  action = g_hash_table_lookup (self->browser_action_map, web_extension);
+  if (!action)
+    return;
+
+  g_assert (g_list_store_find (self->browser_actions, action, &position));
+  g_list_store_remove (self->browser_actions, position);
+
+  g_hash_table_remove (self->browser_action_map, web_extension);
 }
 
 void
@@ -1387,9 +1436,10 @@ ephy_web_extension_manager_set_active (EphyWebExtensionManager *self,
   for (list = windows; list && list->data; list = list->next) {
     EphyWindow *window = EPHY_WINDOW (list->data);
 
-    if (active)
+    if (active) {
       ephy_web_extension_manager_add_web_extension_to_window (self, web_extension, window);
-    else
+      application_window_added_cb (shell, window, web_extension);
+    } else
       ephy_web_extension_manager_remove_web_extension_from_window (self, web_extension, window);
   }
 
@@ -1400,17 +1450,35 @@ ephy_web_extension_manager_set_active (EphyWebExtensionManager *self,
     if (ephy_web_extension_has_background_web_view (web_extension))
       run_background_script (self, web_extension);
 
+    if (ephy_web_extension_has_browser_action (web_extension)) {
+      EphyBrowserAction *action = ephy_browser_action_new (web_extension);
+      g_list_store_append (self->browser_actions, action);
+      g_hash_table_insert (self->browser_action_map, web_extension, g_steal_pointer (&action));
+    }
+
     ephy_web_extension_api_commands_init (web_extension);
   } else {
     g_signal_handlers_disconnect_by_data (shell, web_extension);
 
-    g_hash_table_remove (self->browser_action_map, web_extension);
+    remove_browser_action (self, web_extension);
     g_hash_table_remove (self->background_web_views, web_extension);
     g_object_set_data (G_OBJECT (web_extension), "alarms", NULL); /* Set in alarms.c */
     ephy_web_extension_api_commands_dispose (web_extension);
   }
 }
 
+void
+ephy_web_extension_manager_show_browser_action (EphyWebExtensionManager *self,
+                                                EphyWebExtension        *web_extension)
+{
+  EphyBrowserAction *action = EPHY_BROWSER_ACTION (g_hash_table_lookup (self->browser_action_map, web_extension));
+
+  if (!action || ephy_browser_action_activate (action))
+    return;
+
+  g_signal_emit (self, signals[SHOW_BROWSER_ACTION], 0, action);
+}
+
 GtkWidget *
 ephy_web_extension_manager_get_page_action (EphyWebExtensionManager *self,
                                             EphyWebExtension        *web_extension,
@@ -1426,6 +1494,12 @@ ephy_web_extension_manager_get_page_action (EphyWebExtensionManager *self,
   return ret;
 }
 
+GListStore *
+ephy_web_extension_manager_get_browser_actions (EphyWebExtensionManager *self)
+{
+  return self->browser_actions;
+}
+
 void
 ephy_web_extension_manager_handle_notifications_action (EphyWebExtensionManager *self,
                                                         GVariant                *params)
@@ -1513,16 +1587,16 @@ tab_emit_ready_cb (GObject      *source,
   PendingMessageReplyTracker *tracker = user_data;
   GHashTable *pending_messages;
   g_autoptr (GError) error = NULL;
-  g_autoptr (WebKitJavascriptResult) js_result = NULL;
+  g_autoptr (JSCValue) value = NULL;
   GTask *pending_task;
 
-  js_result = webkit_web_view_run_javascript_finish (WEBKIT_WEB_VIEW (source),
-                                                     result,
-                                                     &error);
+  value = webkit_web_view_evaluate_javascript_finish (WEBKIT_WEB_VIEW (source),
+                                                      result,
+                                                      &error);
 
   /* If it returned true it will be asynchronously handled later. Otherwise we
    * complete it now with undefined. */
-  if (error || !jsc_value_to_boolean (webkit_javascript_result_get_js_value (js_result))) {
+  if (error || !jsc_value_to_boolean (value)) {
     pending_messages = g_hash_table_lookup (manager->pending_messages, tracker->web_extension);
     pending_task = g_hash_table_lookup (pending_messages, tracker->message_guid);
     if (pending_task) {
@@ -1565,12 +1639,13 @@ ephy_web_extension_manager_emit_in_tab_with_reply (EphyWebExtensionManager *self
   tracker = g_new0 (PendingMessageReplyTracker, 1);
   tracker->web_extension = web_extension;
   tracker->message_guid = message_guid;
-  webkit_web_view_run_javascript_in_world (target_web_view,
-                                           script,
-                                           ephy_web_extension_get_guid (web_extension),
-                                           NULL,
-                                           tab_emit_ready_cb,
-                                           tracker);
+  webkit_web_view_evaluate_javascript (target_web_view,
+                                       script, -1,
+                                       ephy_web_extension_get_guid (web_extension),
+                                       NULL,
+                                       NULL,
+                                       tab_emit_ready_cb,
+                                       tracker);
 
   pending_messages = g_hash_table_lookup (self->pending_messages, web_extension);
   if (!pending_messages) {
@@ -1591,13 +1666,13 @@ on_extension_emit_ready (GObject      *source,
   PendingMessageReplyTracker *tracker = user_data;
   GHashTable *pending_messages;
   g_autoptr (GError) error = NULL;
-  g_autoptr (WebKitJavascriptResult) js_result = NULL;
+  g_autoptr (JSCValue) value = NULL;
 
-  js_result = webkit_web_view_run_javascript_finish (WEBKIT_WEB_VIEW (source),
-                                                     result,
-                                                     &error);
+  value = webkit_web_view_evaluate_javascript_finish (WEBKIT_WEB_VIEW (source),
+                                                      result,
+                                                      &error);
 
-  if (!error && jsc_value_to_boolean (webkit_javascript_result_get_js_value (js_result)))
+  if (!error && jsc_value_to_boolean (value))
     tracker->handled = TRUE;
 
   /* Once all views have been notified it will either be handled by one of them, in which case
@@ -1625,6 +1700,26 @@ on_extension_emit_ready (GObject      *source,
     g_warning ("Emitting in view errored: %s", error->message);
 }
 
+void
+ephy_web_extension_manager_emit_in_background_view (EphyWebExtensionManager *self,
+                                                    EphyWebExtension        *web_extension,
+                                                    const char              *name,
+                                                    const char              *json)
+{
+  g_autofree char *script = NULL;
+  WebKitWebView *web_view = ephy_web_extension_manager_get_background_web_view (self, web_extension);
+
+  script = g_strdup_printf ("window.browser.%s._emit(%s);", name, json);
+
+  webkit_web_view_evaluate_javascript (web_view,
+                                       script, -1,
+                                       NULL,
+                                       NULL,
+                                       NULL,
+                                       NULL,
+                                       NULL);
+}
+
 static void
 ephy_web_extension_manager_emit_in_extension_views_internal (EphyWebExtensionManager *self,
                                                              EphyWebExtension        *web_extension,
@@ -1660,11 +1755,11 @@ ephy_web_extension_manager_emit_in_extension_views_internal (EphyWebExtensionMan
 
   if (background_view) {
     if (!sender || (sender->view != background_view)) {
-      webkit_web_view_run_javascript (background_view,
-                                      script,
-                                      NULL,
-                                      reply_task ? on_extension_emit_ready : NULL,
-                                      tracker);
+      webkit_web_view_evaluate_javascript (background_view,
+                                           script, -1,
+                                           NULL, NULL, NULL,
+                                           reply_task ? on_extension_emit_ready : NULL,
+                                           tracker);
       pending_views++;
     }
   }
@@ -1675,11 +1770,11 @@ ephy_web_extension_manager_emit_in_extension_views_internal (EphyWebExtensionMan
       if (!sender || (sender->view == popup_view))
         continue;
 
-      webkit_web_view_run_javascript (popup_view,
-                                      script,
-                                      NULL,
-                                      reply_task ? on_extension_emit_ready : NULL,
-                                      tracker);
+      webkit_web_view_evaluate_javascript (popup_view,
+                                           script, -1,
+                                           NULL, NULL, NULL,
+                                           reply_task ? on_extension_emit_ready : NULL,
+                                           tracker);
       pending_views++;
     }
   }
